from typing import Any, Dict, List, Optional

from ...utils.helpers import _get_and_maybe_create_project, _validate_api_key
from ..logs import Log
from .contexts import *
from .logs import *

# Datasets #
# ---------#


def list_datasets(
    *,
    project: Optional[str] = None,
    prefix: str = "",
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    List all datasets associated with a project and context.

    Args:
        project: Name of the project the datasets belong to.

        prefix: Prefix of the datasets to get.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A list of datasets.
    """
    api_key = _validate_api_key(api_key)
    contexts = get_contexts(
        prefix=f"Datasets/{prefix}",
        project=project,
        api_key=api_key,
    )
    return {
        "/".join(name.split("/")[1:]): description
        for name, description in contexts.items()
    }


def upload_dataset(
    name: str,
    data: List[Any],
    *,
    overwrite: bool = False,
    allow_duplicates: bool = False,
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> List[int]:
    """
    Upload a dataset to the server.

    Args:
        name: Name of the dataset.

        data: Contents of the dataset.

        overwrite: Whether to overwrite the dataset if it already exists.

        allow_duplicates: Whether to allow duplicates in the dataset.

        project: Name of the project the dataset belongs to.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.
    Returns:
        A list all log ids in the dataset.
    """
    api_key = _validate_api_key(api_key)
    project = _get_and_maybe_create_project(project, api_key=api_key)
    context = f"Datasets/{name}"
    log_instances = [isinstance(item, unify.Log) for item in data]
    are_logs = False
    if not allow_duplicates and not overwrite:
        # ToDo: remove this verbose logic once ignore_duplicates is implemented
        if name in unify.list_datasets():
            upstream_dataset = unify.Dataset(
                unify.download_dataset(name, project=project, api_key=api_key),
            )
        else:
            upstream_dataset = unify.Dataset([])
    if any(log_instances):
        assert all(log_instances), "If any items are logs, all items must be logs"
        are_logs = True
        # ToDo: remove this verbose logic once ignore_duplicates is implemented
        if not allow_duplicates and not overwrite:
            data = [l for l in data if l not in upstream_dataset]
    elif not all(isinstance(item, dict) for item in data):
        # ToDo: remove this verbose logic once ignore_duplicates is implemented
        if not allow_duplicates and not overwrite:
            data = [item for item in data if item not in upstream_dataset]
        data = [{"data": item} for item in data]
    if name in unify.list_datasets():
        upstream_ids = get_logs(
            project=project,
            context=context,
            return_ids_only=True,
        )
    else:
        upstream_ids = []
    if not are_logs:
        return upstream_ids + create_logs(
            project=project,
            context=context,
            entries=data,
            mutable=True,
            batched=True,
            # ToDo: uncomment once ignore_duplicates is implemented
            # ignore_duplicates=not allow_duplicates,
        )
    local_ids = [l.id for l in data]
    matching_ids = [id for id in upstream_ids if id in local_ids]
    matching_data = [l.entries for l in data if l.id in matching_ids]
    assert len(matching_data) == len(
        matching_ids,
    ), "matching data and ids must be the same length"
    if matching_data:
        update_logs(
            logs=matching_ids,
            api_key=api_key,
            entries=matching_data,
            overwrite=True,
        )
    if overwrite:
        upstream_only_ids = [id for id in upstream_ids if id not in local_ids]
        if upstream_only_ids:
            delete_logs(
                logs=upstream_only_ids,
                context=context,
                project=project,
                api_key=api_key,
            )
            upstream_ids = [id for id in upstream_ids if id not in upstream_only_ids]
    ids_not_in_dataset = [
        id for id in local_ids if id not in matching_ids and id is not None
    ]
    if ids_not_in_dataset:
        if context not in unify.get_contexts():
            unify.create_context(
                context,
                project=project,
                api_key=api_key,
            )
        unify.add_logs_to_context(
            log_ids=ids_not_in_dataset,
            context=context,
            project=project,
            api_key=api_key,
        )
    local_only_data = [l.entries for l in data if l.id is None]
    if local_only_data:
        return upstream_ids + create_logs(
            project=project,
            context=context,
            entries=local_only_data,
            mutable=True,
            batched=True,
        )
    return upstream_ids + ids_not_in_dataset


def download_dataset(
    name: str,
    *,
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> List[Log]:
    """
    Download a dataset from the server.

    Args:
        name: Name of the dataset.

        project: Name of the project the dataset belongs to.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.
    """
    api_key = _validate_api_key(api_key)
    project = _get_and_maybe_create_project(project, api_key=api_key)
    logs = get_logs(
        project=project,
        context=f"Datasets/{name}",
    )
    return list(reversed(logs))


def delete_dataset(
    name: str,
    *,
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> None:
    """
    Delete a dataset from the server.

    Args:
        name: Name of the dataset.

        project: Name of the project the dataset belongs to.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.
    """
    api_key = _validate_api_key(api_key)
    project = _get_and_maybe_create_project(project, api_key=api_key)
    delete_context(f"Datasets/{name}", project=project, api_key=api_key)


def add_dataset_entries(
    name: str,
    data: List[Any],
    *,
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> List[int]:
    """
    Adds entries to an existing dataset in the server.

    Args:
        name: Name of the dataset.

        contents: Contents to add to the dataset.

        project: Name of the project the dataset belongs to.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.
    Returns:
        A list of the newly added dataset logs.
    """
    api_key = _validate_api_key(api_key)
    project = _get_and_maybe_create_project(
        project,
        api_key=api_key,
        create_if_missing=False,
    )
    if not all(isinstance(item, dict) for item in data):
        data = [{"data": item} for item in data]
    logs = create_logs(
        project=project,
        context=f"Datasets/{name}",
        entries=data,
        mutable=True,
        batched=True,
    )
    return logs



import asyncio
import contextvars
import os
import threading
from typing import Any, List

from tqdm import tqdm
from tqdm.asyncio import tqdm_asyncio

MAP_MODE = "threading"


def set_map_mode(mode: str):
    global MAP_MODE
    MAP_MODE = mode


def get_map_mode() -> str:
    return MAP_MODE


def _is_iterable(item: Any) -> bool:
    try:
        iter(item)
        return True
    except TypeError:
        return False


# noinspection PyShadowingBuiltins
def map(
    fn: callable,
    *args,
    mode=None,
    name="",
    from_args=False,
    raise_exceptions=True,
    **kwargs,
) -> Any:

    if name:
        name = (
            " ".join(substr[0].upper() + substr[1:] for substr in name.split("_")) + " "
        )

    if mode is None:
        mode = get_map_mode()

    assert mode in (
        "threading",
        "asyncio",
        "loop",
    ), "map mode must be one of threading, asyncio or loop."

    def fn_w_exception_handling(*a, **kw):
        try:
            return fn(*a, **kw)
        except Exception as e:
            if raise_exceptions:
                raise e

    if from_args:
        args = list(args)
        for i, a in enumerate(args):
            if _is_iterable(a):
                args[i] = list(a)

        if args:
            num_calls = len(args[0])
        else:
            for v in kwargs.values():
                if isinstance(v, list):
                    num_calls = len(v)
                    break
            else:
                raise Exception(
                    "At least one of the args or kwargs must be a list, "
                    "which is to be mapped across the threads",
                )
        args_n_kwargs = [
            (
                tuple(a[i] for a in args),
                {
                    k: v[i] if (isinstance(v, list) or isinstance(v, tuple)) else v
                    for k, v in kwargs.items()
                },
            )
            for i in range(num_calls)
        ]
    else:
        args_n_kwargs = args[0]
        if not isinstance(args_n_kwargs[0], tuple):
            if isinstance(args_n_kwargs[0], dict):
                args_n_kwargs = [((), item) for item in args_n_kwargs]
            else:
                args_n_kwargs = [((item,), {}) for item in args_n_kwargs]
        elif (
            not isinstance(args_n_kwargs[0][0], tuple)
            or len(args_n_kwargs[0]) < 2
            or not isinstance(args_n_kwargs[0][1], dict)
        ):
            args_n_kwargs = [(item, {}) for item in args_n_kwargs]
        num_calls = len(args_n_kwargs)

    if mode == "loop":

        pbar = tqdm(total=num_calls, disable=os.environ.get("TQDM_DISABLE", "0") == "1")
        pbar.set_description(f"{name}Iterations")

        returns = list()
        for a, kw in args_n_kwargs:
            ret = fn_w_exception_handling(*a, **kw)
            returns.append(ret)
            pbar.update(1)
        pbar.close()
        return returns

    elif mode == "threading":

        pbar = tqdm(total=num_calls, disable=os.environ.get("TQDM_DISABLE", "0") == "1")
        pbar.set_description(f"{name}Threads")

        def fn_w_indexing(rets: List[None], thread_idx: int, *a, **kw):
            for var, value in kw["context"].items():
                var.set(value)
            del kw["context"]
            ret = fn_w_exception_handling(*a, **kw)
            pbar.update(1)
            rets[thread_idx] = ret

        threads = list()
        returns = [None] * num_calls
        for i, a_n_kw in enumerate(args_n_kwargs):
            a, kw = a_n_kw
            kw["context"] = contextvars.copy_context()
            thread = threading.Thread(
                target=fn_w_indexing,
                args=(returns, i, *a),
                kwargs=kw,
            )
            thread.start()
            threads.append(thread)
        [thread.join() for thread in threads]
        pbar.close()
        return returns

    def _run_asyncio_in_thread(ret):
        asyncio.set_event_loop(asyncio.new_event_loop())
        MAX_WORKERS = 100
        semaphore = asyncio.Semaphore(MAX_WORKERS)
        fns = []

        async def fn_wrapper(*args, **kwargs):
            async with semaphore:
                return await asyncio.to_thread(fn_w_exception_handling, *args, **kwargs)

        for _, a_n_kw in enumerate(args_n_kwargs):
            a, kw = a_n_kw
            fns.append(fn_wrapper(*a, **kw))

        async def main(fns):
            return await tqdm_asyncio.gather(
                *fns,
                desc=f"{name}Coroutines",
                disable=os.environ.get("TQDM_DISABLE", "0") == "1",
            )

        ret += asyncio.run(main(fns))

    ret = []
    thread = threading.Thread(target=_run_asyncio_in_thread, args=(ret,))
    thread.start()
    thread.join()
    return ret



import datetime
from typing import Any, Dict, List, Optional, Union

from unify import BASE_URL
from unify.utils import _requests

from ...utils.helpers import _validate_api_key


def get_query_tags(
    *,
    api_key: Optional[str] = None,
) -> List[str]:
    """
    Get a list of available query tags.

    Args:
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A list of available query tags if successful, otherwise an empty list.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/tags"
    response = _requests.get(url, headers=headers)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()


def get_queries(
    *,
    tags: Optional[Union[str, List[str]]] = None,
    endpoints: Optional[Union[str, List[str]]] = None,
    start_time: Optional[Union[datetime.datetime, str]] = None,
    end_time: Optional[Union[datetime.datetime, str]] = None,
    page_number: Optional[int] = None,
    failures: Optional[Union[bool, str]] = None,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Get query history based on specified filters.

    Args:
        tags: Tags to filter for queries that are marked with these tags.

        endpoints: Optionally specify an endpoint, or a list of endpoints to filter for.

        start_time: Timestamp of the earliest query to aggregate.
        Format is `YYYY-MM-DD hh:mm:ss`.

        end_time: Timestamp of the latest query to aggregate.
        Format is `YYYY-MM-DD hh:mm:ss`.

        page_number: The query history is returned in pages, with up to 100 prompts per
        page. Increase the page number to see older prompts. Default is 1.

        failures: indicates whether to includes failures in the return
        (when set as True), or whether to return failures exclusively
        (when set as ‘only’). Default is False.

        api_key: If specified, unify API key to be used.
        Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the query history data.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    params = {}
    if tags:
        params["tags"] = tags
    if endpoints:
        params["endpoints"] = endpoints
    if start_time:
        params["start_time"] = start_time
    if end_time:
        params["end_time"] = end_time
    if page_number:
        params["page_number"] = page_number
    if failures:
        params["failures"] = failures

    url = f"{BASE_URL}/queries"
    response = _requests.get(url, headers=headers, params=params)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()


def log_query(
    *,
    endpoint: str,
    query_body: Dict,
    response_body: Optional[Dict] = None,
    tags: Optional[List[str]] = None,
    timestamp: Optional[Union[datetime.datetime, str]] = None,
    api_key: Optional[str] = None,
):
    """
    Log a query (and optionally response) for a locally deployed (non-Unify-registered)
    model, with tagging (default None) and timestamp (default datetime.now() also
    optionally writeable.

    Args:
        endpoint: Endpoint to log query for.
        query_body: A dict containing the body of the request.
        response_body: An optional dict containing the response to the request.
        tags: Custom tags for later filtering.
        timestamp: A timestamp (if not set, will be the time of sending).
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response message if successful.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    data = {
        "endpoint": endpoint,
        "query_body": query_body,
        "response_body": response_body,
        "tags": tags,
        "timestamp": timestamp,
    }

    # Remove None values from params
    data = {k: v for k, v in data.items() if v is not None}

    url = f"{BASE_URL}/queries"

    response = _requests.post(url, headers=headers, json=data)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()


def get_query_metrics(
    *,
    start_time: Optional[Union[datetime.datetime, str]] = None,
    end_time: Optional[Union[datetime.datetime, str]] = None,
    models: Optional[str] = None,
    providers: Optional[str] = None,
    interval: int = 300,
    secondary_user_id: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Get query metrics for specified parameters.

    Args:
        start_time: Timestamp of the earliest query to aggregate. Format is `YYYY-MM-DD hh:mm:ss`.
        end_time: Timestamp of the latest query to aggregate. Format is `YYYY-MM-DD hh:mm:ss`.
        models: Models to fetch metrics from. Comma-separated string of model names.
        providers: Providers to fetch metrics from. Comma-separated string of provider names.
        interval: Number of seconds in the aggregation interval. Default is 300.
        secondary_user_id: Secondary user id to match the `user` attribute from `/chat/completions`.
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the query metrics.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    params = {
        "start_time": start_time,
        "end_time": end_time,
        "models": models,
        "providers": providers,
        "interval": interval,
        "secondary_user_id": secondary_user_id,
    }

    # Remove None values from params
    params = {k: v for k, v in params.items() if v is not None}

    url = f"{BASE_URL}/metrics"

    response = _requests.get(url, headers=headers, params=params)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()



from typing import Any, Dict, List, Optional

from unify import BASE_URL
from unify.utils import _requests

from ...utils.helpers import _validate_api_key


def create_custom_endpoint(
    *,
    name: str,
    url: str,
    key_name: str,
    model_name: Optional[str] = None,
    provider: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Create a custom endpoint for API calls.

    Args:
        name: Alias for the custom endpoint. This will be the name used to call the endpoint.
        url: Base URL of the endpoint being called. Must support the OpenAI format.
        key_name: Name of the API key that will be passed as part of the query.
        model_name: Name passed to the custom endpoint as model name. If not specified, it will default to the endpoint alias.
        provider: If the custom endpoint is for a fine-tuned model which is hosted directly via one of the supported providers,
                  then this argument should be specified as the provider used.
        api_key: If specified, unify API key to be used. Defaults to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response from the API.

    Raises:
        requests.HTTPError: If the API request fails.
        KeyError: If the UNIFY_KEY is not set and no api_key is provided.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    params = {
        "name": name,
        "url": url,
        "key_name": key_name,
    }

    if model_name:
        params["model_name"] = model_name
    if provider:
        params["provider"] = provider

    response = _requests.post(
        f"{BASE_URL}/custom_endpoint",
        headers=headers,
        params=params,
    )
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()


def delete_custom_endpoint(
    name: str,
    *,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Delete a custom endpoint.

    Args:
        name: Name of the custom endpoint to delete.
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response message.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/custom_endpoint"

    params = {"name": name}

    response = _requests.delete(url, headers=headers, params=params)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()


def rename_custom_endpoint(
    name: str,
    new_name: str,
    *,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Rename a custom endpoint.

    Args:
        name: Name of the custom endpoint to be updated.
        new_name: New name for the custom endpoint.
        api_key: If specified, unify API key to be used. Defaults
                 to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A dictionary containing the response information.

    Raises:
        requests.HTTPError: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/custom_endpoint/rename"

    params = {"name": name, "new_name": new_name}

    response = _requests.post(url, headers=headers, params=params)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()


def list_custom_endpoints(
    *,
    api_key: Optional[str] = None,
) -> List[Dict[str, str]]:
    """
    Get a list of custom endpoints for the authenticated user.

    Args:
        api_key: If specified, unify API key to be used. Defaults
        to the value in the `UNIFY_KEY` environment variable.

    Returns:
        A list of dictionaries containing information about custom endpoints.
        Each dictionary has keys: 'name', 'mdl_name', 'url', and 'key'.

    Raises:
        requests.exceptions.RequestException: If the API request fails.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    url = f"{BASE_URL}/custom_endpoint/list"

    response = _requests.get(url, headers=headers)
    if response.status_code != 200:
        raise Exception(response.json())

    return response.json()



"""
Main caching module providing high-level caching functionality.

This module provides decorators and utilities for caching function results
with multiple backend options and flexible caching modes.
"""

import difflib
import inspect
import json
import threading
from typing import Any, Dict, Optional, Type, Union

from openai.types.chat import ChatCompletion, ParsedChatCompletion
from pydantic import BaseModel
from unify.utils.caching import BaseCache, LocalCache, LocalSeparateCache, RemoteCache
from unify.utils.caching.cache_benchmark import record_get_cache, record_write_to_cache

# Global state
CACHE_LOCK = threading.Lock()
CACHING_ENABLED = False
CURRENT_CACHE_BACKEND = "local"

# Available cache backends
CACHE_BACKENDS = {
    "local": LocalCache,
    "remote": RemoteCache,
    "local_separate": LocalSeparateCache,
}


def set_cache_backend(backend: str) -> None:
    """Set the current cache backend."""
    global CURRENT_CACHE_BACKEND
    if backend not in CACHE_BACKENDS:
        raise ValueError(
            f"Invalid backend: {backend}. Available: {list(CACHE_BACKENDS.keys())}",
        )
    CURRENT_CACHE_BACKEND = backend


def get_cache_backend(backend: Optional[str] = None) -> Type[BaseCache]:
    """Get the cache backend class."""
    if backend is None:
        backend = CURRENT_CACHE_BACKEND
    if backend not in CACHE_BACKENDS:
        raise ValueError(
            f"Invalid backend: {backend}. Available: {list(CACHE_BACKENDS.keys())}",
        )
    return CACHE_BACKENDS[backend]


def set_caching(value: bool) -> bool:
    """Enable or disable caching globally."""
    global CACHING_ENABLED
    CACHING_ENABLED = value
    return CACHING_ENABLED


def is_caching_enabled() -> bool:
    """Check if caching is globally enabled."""
    return CACHING_ENABLED


def _minimal_char_diff(a: str, b: str, context: int = 5) -> str:
    matcher = difflib.SequenceMatcher(None, a, b)
    diff_parts = []

    for tag, i1, i2, j1, j2 in matcher.get_opcodes():
        if tag == "equal":
            segment = a[i1:i2]
            # If the segment is too long, show only a context at the beginning and end.
            if len(segment) > 2 * context:
                diff_parts.append(segment[:context] + "..." + segment[-context:])
            else:
                diff_parts.append(segment)
        elif tag == "replace":
            diff_parts.append(f"[{a[i1:i2]}|{b[j1:j2]}]")
        elif tag == "delete":
            diff_parts.append(f"[-{a[i1:i2]}-]")
        elif tag == "insert":
            diff_parts.append(f"[+{b[j1:j2]}+]")

    return "".join(diff_parts)


@record_get_cache
def _get_cache(
    fn_name: str,
    kw: Dict[str, Any],
    filename: str = None,
    raise_on_empty: bool = False,
    read_closest: bool = False,
    delete_closest: bool = False,
    backend: Optional[str] = None,
) -> Optional[Any]:
    global CACHE_LOCK
    # Prevents circular import
    from unify.logging.logs import Log

    type_mapping = {
        "ChatCompletion": ChatCompletion,
        "Log": Log,
        "ParsedChatCompletion": ParsedChatCompletion,
    }
    CACHE_LOCK.acquire()
    try:
        current_backend = get_cache_backend(backend)
        current_backend.initialize_cache(filename)
        kw = {k: v for k, v in kw.items() if v is not None}
        kw_str = BaseCache.serialize_object(kw)
        cache_str = f"{fn_name}_{kw_str}"
        if not current_backend.has_key(cache_str):
            if raise_on_empty or read_closest:
                keys_to_search = current_backend.list_keys()
                if len(keys_to_search) == 0:
                    CACHE_LOCK.release()
                    raise Exception(
                        f"Failed to get cache for function {fn_name} with kwargs {BaseCache.serialize_object(kw, indent=4)} "
                        f"Cache is empty, mode is read-only ",
                    )
                closest_match = difflib.get_close_matches(
                    cache_str,
                    keys_to_search,
                    n=1,
                    cutoff=0,
                )[0]
                minimal_char_diff = _minimal_char_diff(cache_str, closest_match)
                if read_closest:
                    cache_str = closest_match
                else:
                    CACHE_LOCK.release()
                    raise Exception(
                        f"Failed to get cache for function {fn_name} with kwargs {BaseCache.serialize_object(kw, indent=4)} "
                        f"from cache at {filename}. \n\nCorresponding key\n{cache_str}\nwas not found in the cache.\n\n"
                        f"The closest match is:\n{closest_match}\n\n"
                        f"The contracted diff is:\n{minimal_char_diff}\n\n",
                    )
            else:
                CACHE_LOCK.release()
                return
        ret, res_types = current_backend.retrieve_entry(cache_str)
        if res_types is None:
            CACHE_LOCK.release()
            return ret
        for idx_str, type_str in res_types.items():
            type_str = type_str.split("[")[0]
            idx_list = json.loads(idx_str)
            if len(idx_list) == 0:
                if read_closest and delete_closest:
                    current_backend.remove_entry(cache_str)
                CACHE_LOCK.release()
                typ = type_mapping[type_str]
                if issubclass(typ, BaseModel):
                    return typ(**ret)
                elif issubclass(typ, Log):
                    return typ.from_json(ret)
                raise Exception(f"Cache indexing found for unsupported type: {typ}")
            item = ret
            for i, idx in enumerate(idx_list):
                if i == len(idx_list) - 1:
                    typ = type_mapping[type_str]
                    if issubclass(typ, BaseModel) or issubclass(typ, Log):
                        item[idx] = typ.from_json(item[idx])
                    else:
                        raise Exception(
                            f"Cache indexing found for unsupported type: {typ}",
                        )
                    break
                item = item[idx]
        if read_closest and delete_closest:
            current_backend.remove_entry(cache_str)
        CACHE_LOCK.release()
        return ret
    except Exception as e:
        if CACHE_LOCK.locked():
            CACHE_LOCK.release()
        raise Exception(
            f"Failed to get cache for function {fn_name} with kwargs {kw} "
            f"from cache at {filename}",
        ) from e


@record_write_to_cache
def _write_to_cache(
    fn_name: str,
    kw: Dict[str, Any],
    response: Any,
    backend: Optional[str] = None,
    filename: str = None,
):

    global CACHE_LOCK
    CACHE_LOCK.acquire()
    try:
        current_backend = get_cache_backend(backend)
        current_backend.initialize_cache(filename)
        kw = {k: v for k, v in kw.items() if v is not None}
        kw_str = BaseCache.serialize_object(kw)
        cache_str = f"{fn_name}_{kw_str}"
        res_types = {}
        response_str = BaseCache.serialize_object(response, res_types)
        current_backend.store_entry(
            key=cache_str,
            value=response_str,
            res_types=res_types,
        )
        CACHE_LOCK.release()
    except Exception as e:
        CACHE_LOCK.release()
        raise Exception(
            f"Failed to write function {fn_name} with kwargs {kw} and "
            f"response {response} to cache at {filename}",
        ) from e


def _handle_reading_from_cache(
    fn_name: str,
    kwargs: Dict[str, Any],
    mode: str,
    backend: Optional[str] = None,
):
    if isinstance(mode, str) and mode.endswith("-closest"):
        mode = mode.removesuffix("-closest")
        read_closest = True
    else:
        read_closest = False
    in_cache = False
    ret = None
    if mode in [True, "both", "read", "read-only"]:
        ret = _get_cache(
            fn_name=fn_name,
            kw=kwargs,
            raise_on_empty=mode == "read-only",
            read_closest=read_closest,
            delete_closest=read_closest,
            backend=backend,
        )
        in_cache = True if ret is not None else False
    return ret, read_closest, in_cache


# Decorators #
# -----------#


def cached(
    fn: callable = None,
    *,
    mode: Union[bool, str] = True,
    backend: Optional[str] = None,
):
    if fn is None:
        return lambda f: cached(
            f,
            mode=mode,
            backend=backend,
        )

    def wrapped(*args, **kwargs):
        sig = inspect.signature(fn)
        bound = sig.bind_partial(*args, **kwargs)
        bound.apply_defaults()
        args_kwargs = bound.arguments
        ret, read_closest, in_cache = _handle_reading_from_cache(
            fn.__name__,
            args_kwargs,
            mode,
            backend,
        )
        if ret is None:
            ret = fn(*args, **kwargs)
        if (ret is not None or read_closest) and mode in [
            True,
            "both",
            "write",
        ]:
            if not in_cache or mode == "write":
                _write_to_cache(
                    fn_name=fn.__name__,
                    kw=args_kwargs,
                    response=ret,
                    backend=backend,
                )
        return ret

    async def async_wrapped(*args, **kwargs):
        sig = inspect.signature(fn)
        bound = sig.bind_partial(*args, **kwargs)
        bound.apply_defaults()
        args_kwargs = bound.arguments
        ret, read_closest, in_cache = _handle_reading_from_cache(
            fn.__name__,
            args_kwargs,
            mode,
            backend,
        )
        if ret is None:
            ret = await fn(*args, **kwargs)
        if (ret is not None or read_closest) and mode in [
            True,
            "both",
            "write",
        ]:
            if not in_cache or mode == "write":
                _write_to_cache(
                    fn_name=fn.__name__,
                    kw=args_kwargs,
                    response=ret,
                    backend=backend,
                )
        return ret

    return wrapped if not inspect.iscoroutinefunction(fn) else async_wrapped


# File Manipulation #
# ------------------#


def cache_file_union(
    first_cache_fpath: str,
    second_cache_fpath: str,
    target_cache_fpath: str,
    conflict_mode="raise",
):
    with open(first_cache_fpath, "r") as file:
        first_cache = json.load(file)
    with open(second_cache_fpath, "r") as file:
        second_cache = json.load(file)
    if conflict_mode == "raise":
        for key, value in first_cache.items():
            if key in second_cache:
                assert second_cache[key] == value, (
                    f"key {key} found in both caches, but values conflict:"
                    f"{first_cache_fpath} had value: {value}"
                    f"{second_cache_fpath} had value: {second_cache[key]}"
                )
        union_cache = {**first_cache, **second_cache}
    elif conflict_mode == "first_overrides":
        union_cache = {**second_cache, **first_cache}
    elif conflict_mode == "second_overrides":
        union_cache = {**first_cache, **second_cache}
    else:
        raise Exception(
            "Invalud conflict_mode, must be one of: 'raise', 'first_overrides' or 'second_overrides'",
        )
    with open(target_cache_fpath, "w+") as file:
        json.dump(union_cache, file)


def cache_file_intersection(
    first_cache_fpath: str,
    second_cache_fpath: str,
    target_cache_fpath: str,
    conflict_mode="raise",
):
    with open(first_cache_fpath, "r") as file:
        first_cache = json.load(file)
    with open(second_cache_fpath, "r") as file:
        second_cache = json.load(file)
    if conflict_mode == "raise":
        for key, value in first_cache.items():
            if key in second_cache:
                assert second_cache[key] == value, (
                    f"key {key} found in both caches, but values conflict:"
                    f"{first_cache_fpath} had value: {value}"
                    f"{second_cache_fpath} had value: {second_cache[key]}"
                )
        intersection_cache = {k: v for k, v in first_cache.items() if k in second_cache}
    elif conflict_mode == "first_overrides":
        intersection_cache = {k: v for k, v in first_cache.items() if k in second_cache}
    elif conflict_mode == "second_overrides":
        intersection_cache = {k: v for k, v in second_cache.items() if k in first_cache}
    else:
        raise Exception(
            "Invalud conflict_mode, must be one of: 'raise', 'first_overrides' or 'second_overrides'",
        )
    with open(target_cache_fpath, "w+") as file:
        json.dump(intersection_cache, file)


def subtract_cache_files(
    first_cache_fpath: str,
    second_cache_fpath: str,
    target_cache_fpath: str,
    raise_on_conflict=True,
):
    with open(first_cache_fpath, "r") as file:
        first_cache = json.load(file)
    with open(second_cache_fpath, "r") as file:
        second_cache = json.load(file)
    if raise_on_conflict:
        for key, value in first_cache.items():
            if key in second_cache:
                assert second_cache[key] == value, (
                    f"key {key} found in both caches, but values conflict:"
                    f"{first_cache_fpath} had value: {value}"
                    f"{second_cache_fpath} had value: {second_cache[key]}"
                )
    final_cache = {k: v for k, v in first_cache.items() if k not in second_cache}
    with open(target_cache_fpath, "w+") as file:
        json.dump(final_cache, file)



"""
Remote cache implementation using the logging system.

This cache stores data remotely using the Unify logging infrastructure.
"""

import json
from typing import Any, Dict, List, Optional

from unify.utils.caching.base_cache import BaseCache


class RemoteCache(BaseCache):
    """Remote cache implementation using the logging system."""

    _remote_context = "UNIFY_CACHE"

    @staticmethod
    def _build_filter_expression(cache_key: str) -> str:
        """Build a filter expression for querying logs."""
        return f"key == {json.dumps(cache_key)}"

    @classmethod
    def set_cache_name(cls, name: str) -> None:
        """Set the remote context name for the cache."""
        cls._remote_context = name

    @classmethod
    def get_cache_name(cls) -> str:
        """Get the current remote context name."""
        return cls._remote_context

    @classmethod
    def store_entry(
        cls,
        *,
        key: str,
        value: Any,
        res_types: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Store a key-value pair in the remote cache."""
        from unify import delete_logs, get_logs, log

        # Remove existing entries with the same key
        existing_logs = get_logs(
            context=cls._remote_context,
            filter=cls._build_filter_expression(key),
            return_ids_only=True,
        )

        if existing_logs:
            delete_logs(logs=existing_logs, context=cls._remote_context)

        # Create new log entry
        entries = {"value": value}
        if res_types:
            entries["res_types"] = json.dumps(res_types)
        log(key=key, context=cls._remote_context, **entries)

    @classmethod
    def write(cls, filename: str = None) -> None:
        """No-op for remote cache - data is persisted immediately."""

    @classmethod
    def initialize_cache(cls, name: str = None) -> None:
        """Ensure the remote context exists."""
        from unify import create_context, get_contexts

        if cls._remote_context not in get_contexts():
            create_context(cls._remote_context)

    @classmethod
    def list_keys(cls) -> List[str]:
        """Get a list of all cache keys from the remote context."""
        from unify import get_logs

        logs = get_logs(context=cls._remote_context)
        return [log.entries["key"] for log in logs]

    @classmethod
    def retrieve_entry(cls, key: str) -> tuple[Optional[Any], Optional[Dict[str, Any]]]:
        """
        Retrieve a value from the remote cache.

        Returns:
            Tuple of (value, res_types) or (None, None) if not found
        """
        from unify import get_logs

        logs = get_logs(
            context=cls._remote_context,
            filter=cls._build_filter_expression(key),
        )

        if not logs:
            return None, None

        entry = logs[0].entries
        value = json.loads(entry["value"])
        res_types = None

        if "res_types" in entry:
            res_types = json.loads(entry["res_types"])

        return value, res_types

    @classmethod
    def has_key(cls, key: str) -> bool:
        """Check if a key exists in the remote cache."""
        from unify import get_logs

        logs = get_logs(
            context=cls._remote_context,
            filter=cls._build_filter_expression(key),
            return_ids_only=True,
        )
        return len(logs) > 0

    @classmethod
    def remove_entry(cls, key: str) -> None:
        """Remove an entry from the remote cache."""
        from unify import delete_logs, get_logs

        logs = get_logs(
            context=cls._remote_context,
            filter=cls._build_filter_expression(key),
            return_ids_only=True,
        )
        if logs:
            delete_logs(context=cls._remote_context, logs=logs)



"""
Abstract base class for cache implementations.

This class defines the interface that all cache backends must implement.
"""

import inspect
import json
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel


class BaseCache(ABC):
    """Abstract base class for cache implementations."""

    @staticmethod
    def serialize_object(
        obj: Any,
        cached_types: Dict[str, str] = None,
        idx: List[Union[str, int]] = None,
        indent: int = None,
    ) -> Any:
        """
        Serialize an object to a JSON-serializable format.

        Args:
            obj: Object to serialize
            type_registry: Dictionary to track object types for reconstruction
            path: Current path in the object structure
            indent: JSON indentation level

        Returns:
            Serialized object or JSON string if at root level
        """
        # Prevent circular import
        from unify.logging.logs import Log

        base = False
        if idx is None:
            base = True
            idx = list()
        if isinstance(obj, BaseModel):
            if cached_types is not None:
                cached_types[json.dumps(idx, indent=indent)] = obj.__class__.__name__
            ret = obj.model_dump()
        elif inspect.isclass(obj) and issubclass(obj, BaseModel):
            ret = obj.schema_json()
        elif isinstance(obj, Log):
            if cached_types is not None:
                cached_types[json.dumps(idx, indent=indent)] = obj.__class__.__name__
            ret = obj.to_json()
        elif isinstance(obj, dict):
            ret = {
                k: BaseCache.serialize_object(v, cached_types, idx + ["k"])
                for k, v in obj.items()
            }
        elif isinstance(obj, list):
            ret = [
                BaseCache.serialize_object(v, cached_types, idx + [i])
                for i, v in enumerate(obj)
            ]
        elif isinstance(obj, tuple):
            ret = tuple(
                BaseCache.serialize_object(v, cached_types, idx + [i])
                for i, v in enumerate(obj)
            )
        else:
            ret = obj
        return json.dumps(ret, indent=indent) if base else ret

    @classmethod
    @abstractmethod
    def set_cache_name(cls, name: str) -> None:
        """Set the cache identifier/name."""

    @classmethod
    @abstractmethod
    def get_cache_name(cls) -> str:
        """Get the current cache identifier/name."""

    @classmethod
    @abstractmethod
    def store_entry(
        cls,
        *,
        key: str,
        value: Any,
        res_types: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Store a key-value pair in the cache."""

    @classmethod
    @abstractmethod
    def write(cls, filename: str = None) -> None:
        pass

    @classmethod
    @abstractmethod
    def initialize_cache(cls, name: str = None) -> None:
        """Initialize or load the cache from storage."""

    @classmethod
    @abstractmethod
    def list_keys(cls) -> List[str]:
        """Get a list of all cache keys."""

    @classmethod
    @abstractmethod
    def retrieve_entry(cls, key: str) -> tuple[Optional[Any], Optional[Dict[str, Any]]]:
        """
        Retrieve a value from the cache.

        Returns:
            Tuple of (value, res_types) or (None, None) if not found
        """

    @classmethod
    @abstractmethod
    def has_key(cls, key: str) -> bool:
        """Check if a key exists in the cache."""

    @classmethod
    @abstractmethod
    def remove_entry(cls, key: str) -> None:
        """Remove an entry from the cache."""



import json
import os
from typing import Optional
from urllib.parse import urlparse

import httpx


def _unify_requests_debug_enabled() -> bool:
    return os.getenv("UNIFY_REQUESTS_DEBUG", "false").lower() in ("true", "1")


def make_httpx_client_for_unify_logging(base_url: str) -> Optional[httpx.Client]:
    if not _unify_requests_debug_enabled():
        return None

    from unify.utils import _requests as _unify_requests

    parsed = urlparse(base_url)
    base_host = parsed.hostname
    base_scheme = parsed.scheme

    def _is_unify_chat_request(request: httpx.Request) -> bool:
        try:
            if request.url.host != base_host or request.url.scheme != base_scheme:
                return False
            return request.url.path.endswith("/chat/completions")
        except Exception:
            return False

    def _pre_request_log(request: httpx.Request) -> None:
        try:
            if not _is_unify_chat_request(request):
                return
            method = request.method.upper()
            url_str = str(request.url)

            headers = dict(request.headers)
            # normalize Authorization header key for masking in _requests._log
            auth_val = headers.get("Authorization", headers.get("authorization"))
            if auth_val is not None:
                headers["Authorization"] = auth_val
                if "authorization" in headers:
                    try:
                        del headers["authorization"]
                    except Exception:
                        pass
            params = dict(request.url.params)
            body_json = None
            if request.content:
                try:
                    body_json = json.loads(
                        (
                            request.content.decode("utf-8")
                            if isinstance(request.content, (bytes, bytearray))
                            else request.content
                        ),
                    )
                except Exception:
                    body_json = None

            kw = {"headers": headers}
            if params:
                kw["params"] = params
            if body_json is not None:
                kw["json"] = body_json

            _unify_requests._log(method, url_str, True, **kw)
        except Exception:
            pass

    def _post_response_log(response: httpx.Response) -> None:
        try:
            request = response.request
            if not _is_unify_chat_request(request):
                return
            method = request.method.upper()
            url_str = str(request.url)

            is_stream = False
            try:
                req_body = None
                if request.content:
                    try:
                        req_body = json.loads(
                            (
                                request.content.decode("utf-8")
                                if isinstance(request.content, (bytes, bytearray))
                                else request.content
                            ),
                        )
                    except Exception:
                        req_body = None
                if isinstance(req_body, dict) and req_body.get("stream") is True:
                    is_stream = True
                if "text/event-stream" in (
                    response.headers.get("content-type", "").lower()
                ):
                    is_stream = True
            except Exception:
                pass

            if is_stream:
                _unify_requests._log(
                    f"{method} response:{response.status_code}",
                    url_str,
                    True,
                    response={},
                )
                return

            try:
                response.read()
            except Exception:
                pass
            try:
                payload = response.json()
            except Exception:
                return
            _unify_requests._log(
                f"{method} response:{response.status_code}",
                url_str,
                response=payload,
            )
        except Exception:
            pass

    return httpx.Client(
        event_hooks={"request": [_pre_request_log], "response": [_post_response_log]},
    )


def make_async_httpx_client_for_unify_logging(
    base_url: str,
) -> Optional[httpx.AsyncClient]:
    if not _unify_requests_debug_enabled():
        return None

    from unify.utils import _requests as _unify_requests

    parsed = urlparse(base_url)
    base_host = parsed.hostname
    base_scheme = parsed.scheme

    def _is_unify_chat_request(request: httpx.Request) -> bool:
        try:
            if request.url.host != base_host or request.url.scheme != base_scheme:
                return False
            return request.url.path.endswith("/chat/completions")
        except Exception:
            return False

    async def _pre_request_log(request: httpx.Request) -> None:
        try:
            if not _is_unify_chat_request(request):
                return
            method = request.method.upper()
            url_str = str(request.url)

            headers = dict(request.headers)
            auth_val = headers.get("Authorization", headers.get("authorization"))
            if auth_val is not None:
                headers["Authorization"] = auth_val
                if "authorization" in headers:
                    try:
                        del headers["authorization"]
                    except Exception:
                        pass
            params = dict(request.url.params)
            body_json = None
            if request.content:
                try:
                    body_json = json.loads(
                        (
                            request.content.decode("utf-8")
                            if isinstance(request.content, (bytes, bytearray))
                            else request.content
                        ),
                    )
                except Exception:
                    body_json = None

            kw = {"headers": headers}
            if params:
                kw["params"] = params
            if body_json is not None:
                kw["json"] = body_json

            _unify_requests._log(method, url_str, True, **kw)
        except Exception:
            pass

    async def _post_response_log(response: httpx.Response) -> None:
        try:
            request = response.request
            if not _is_unify_chat_request(request):
                return
            method = request.method.upper()
            url_str = str(request.url)

            is_stream = False
            try:
                req_body = None
                if request.content:
                    try:
                        req_body = json.loads(
                            (
                                request.content.decode("utf-8")
                                if isinstance(request.content, (bytes, bytearray))
                                else request.content
                            ),
                        )
                    except Exception:
                        req_body = None
                if isinstance(req_body, dict) and req_body.get("stream") is True:
                    is_stream = True
                if "text/event-stream" in (
                    response.headers.get("content-type", "").lower()
                ):
                    is_stream = True
            except Exception:
                pass

            if is_stream:
                _unify_requests._log(
                    f"{method} response:{response.status_code}",
                    url_str,
                    True,
                    response={},
                )
                return

            try:
                await response.aread()
            except Exception:
                pass
            try:
                payload = response.json()
            except Exception:
                return
            _unify_requests._log(
                f"{method} response:{response.status_code}",
                url_str,
                response=payload,
            )
        except Exception:
            pass

    return httpx.AsyncClient(
        event_hooks={"request": [_pre_request_log], "response": [_post_response_log]},
    )



"""
Caching utilities for the Unify framework.

This module provides a flexible caching system with multiple backends:
- LocalCache: Simple local file-based caching
- LocalSeparateCache: Separate read/write caches for better performance
- RemoteCache: Remote caching using the logging system
"""

from .base_cache import BaseCache
from .local_cache import LocalCache
from .local_separate_cache import LocalSeparateCache
from .remote_cache import RemoteCache
from .cache_benchmark import CacheStats, get_cache_stats, reset_cache_stats

__all__ = [
    "BaseCache",
    "LocalCache",
    "LocalSeparateCache",
    "RemoteCache",
    "CacheStats",
    "get_cache_stats",
    "reset_cache_stats",
]



import json
import os
import warnings
from typing import Any, Dict, List, Optional

from unify.utils.caching.base_cache import BaseCache


class LocalSeparateCache(BaseCache):
    """Local cache with separate read and write storage for better performance."""

    _cache_read: Optional[Dict[str, Any]] = None
    _cache_write: Optional[Dict[str, Any]] = None
    _cache_dir: str = os.environ.get("UNIFY_CACHE_DIR", os.getcwd())
    _cache_name_read: str = ".cache.json"
    _cache_name_write: str = ".cache_write.json"
    _enabled: bool = False

    @classmethod
    def set_cache_name(cls, name: str) -> None:
        """Set the cache names and reset both caches."""
        cls._cache_name_read = f"{name}_read"
        cls._cache_name_write = f"{name}_write"
        cls._cache_read = None  # Force reload of read cache
        cls._cache_write = None  # Force reload of write cache

    @classmethod
    def get_cache_name(cls) -> str:
        """Get the current read cache name."""
        return cls._cache_name_read

    @classmethod
    def get_cache_filepath(cls, name: str = None) -> str:
        """Get the full filepath for the cache file."""
        if name is None:
            name = cls.get_cache_name()
        return os.path.join(cls._cache_dir, name)

    @classmethod
    def is_enabled(cls) -> bool:
        """Check if the cache is enabled."""
        return cls._enabled

    @classmethod
    def store_entry(
        cls,
        *,
        key: str,
        value: Any,
        res_types: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Store a key-value pair in the write cache."""
        if res_types:
            cls._cache_write[f"{key}_res_types"] = res_types
        cls._cache_write[key] = value
        cls.write()

    @classmethod
    def write(cls, filename: str = None) -> None:
        """Write the write cache to disk."""
        if cls._cache_write is None:
            return

        cache_filepath = os.path.join(cls._cache_dir, cls._cache_name_write)
        with open(cache_filepath, "w") as outfile:
            json.dump(cls._cache_write, outfile, indent=2)

    @classmethod
    def initialize_cache(cls, name: str = None) -> None:
        """Initialize both read and write caches."""
        # Always initialize the write cache
        if cls._cache_write is None:
            cls._cache_write = {}
            write_filepath = os.path.join(cls._cache_dir, cls._cache_name_write)
            with open(write_filepath, "w") as outfile:
                json.dump({}, outfile)

        # Initialize the read cache
        if cls._cache_read is None:
            cls._cache_read = {}
            read_filepath = os.path.join(cls._cache_dir, cls._cache_name_read)

            try:
                if os.path.exists(read_filepath):
                    with open(read_filepath, "r") as infile:
                        content = infile.read().strip()
                        if content:
                            cls._cache_read = json.loads(content)
            except (json.JSONDecodeError, IOError):
                # File contains invalid JSON or can't be read, keep empty cache
                warnings.warn(
                    f"Cache file {read_filepath} contains invalid JSON or can't be read, reinitializing.",
                )
                cls._cache_read = {}

    @classmethod
    def list_keys(cls) -> List[str]:
        return list(cls._cache_read.keys()) + list(cls._cache_write.keys())

    @classmethod
    def retrieve_entry(cls, key: str) -> tuple[Optional[Any], Optional[Dict[str, Any]]]:
        """
        Retrieve a value from the cache, checking write cache first.

        Returns:
            Tuple of (value, res_types) or (None, None) if not found
        """
        # First check the write cache
        if cls._cache_write and key in cls._cache_write:
            value = cls._cache_write[key]
            res_types = cls._cache_write.get(f"{key}_res_types")
            deserialized_value = json.loads(value)
            return deserialized_value, res_types

        # If not found in write cache, check the read cache
        if cls._cache_read and key in cls._cache_read:
            value = cls._cache_read[key]
            res_types = cls._cache_read.get(f"{key}_res_types")

            deserialized_value = json.loads(value)
            # Promote to write cache for faster future access
            cls.store_entry(
                key=key,
                value=cls.serialize_object(deserialized_value),
                res_types=res_types,
            )
            return deserialized_value, res_types

        return None, None

    @classmethod
    def has_key(cls, key: str) -> bool:
        """Check if a key exists in either cache."""
        return (cls._cache_write is not None and key in cls._cache_write) or (
            cls._cache_read is not None and key in cls._cache_read
        )

    @classmethod
    def remove_entry(cls, key: str) -> None:
        """Remove an entry from both caches."""
        if cls._cache_write:
            cls._cache_write.pop(key, None)
            cls._cache_write.pop(f"{key}_res_types", None)
        if cls._cache_read:
            cls._cache_read.pop(key, None)
            cls._cache_read.pop(f"{key}_res_types", None)
        cls.write()



# global
import abc
import asyncio
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union

import requests

# local
import unify

# noinspection PyProtectedMember
from openai._types import Headers, Query
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionMessageParam,
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)
from pydantic import BaseModel
from typing_extensions import Self
from unify import BASE_URL
from unify.utils import _requests

# noinspection PyProtectedMember
from unify.utils.helpers import _default, _validate_api_key

from ..clients import AsyncUnify, _Client, _UniClient


class _MultiClient(_Client, abc.ABC):
    def __init__(
        self,
        endpoints: Optional[Union[str, Iterable[str]]] = None,
        *,
        system_message: Optional[str] = None,
        messages: Optional[
            Union[
                List[ChatCompletionMessageParam],
                Dict[str, List[ChatCompletionMessageParam]],
            ]
        ] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, int]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        max_completion_tokens: Optional[int] = None,
        n: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        response_format: Optional[Union[Type[BaseModel], Dict[str, str]]] = None,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = None,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = None,
        service_tier: Optional[str] = None,
        tools: Optional[Iterable[ChatCompletionToolParam]] = None,
        tool_choice: Optional[ChatCompletionToolChoiceOptionParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        reasoning_effort: Optional[str] = None,
        # platform arguments
        use_custom_keys: bool = False,
        tags: Optional[List[str]] = None,
        drop_params: Optional[bool] = True,
        region: Optional[str] = None,
        log_query_body: Optional[bool] = True,
        log_response_body: Optional[bool] = True,
        api_key: Optional[str] = None,
        # python client arguments
        stateful: bool = False,
        return_full_completion: bool = False,
        traced: bool = False,
        cache: Union[bool, str] = None,
        cache_backend: Optional[str] = None,
        # passthrough arguments
        extra_headers: Optional[Headers] = None,
        extra_query: Optional[Query] = None,
        **kwargs,
    ) -> None:
        """Initialize the Multi LLM Unify client.

        Args:
            endpoints: A single endpoint name or a list of endpoint names, with each name
            in OpenAI API format: <model_name>@<provider_name>. Defaults to None.

            system_message: An optional string containing the system message. This
            always appears at the beginning of the list of messages.

            messages: A list of messages comprising the conversation so far. This will
            be appended to the system_message if it is not None, and any user_message
            will be appended if it is not None.

            frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on their existing frequency in the text so far, decreasing the
            model's likelihood to repeat the same line verbatim.

            logit_bias: Modify the likelihood of specified tokens appearing in the
            completion. Accepts a JSON object that maps tokens (specified by their token
            ID in the tokenizer) to an associated bias value from -100 to 100.
            Mathematically, the bias is added to the logits generated by the model prior
            to sampling. The exact effect will vary per model, but values between -1 and
            1 should decrease or increase likelihood of selection; values like -100 or
            100 should result in a ban or exclusive selection of the relevant token.

            logprobs: Whether to return log probabilities of the output tokens or not.
            If true, returns the log probabilities of each output token returned in the
            content of message.

            top_logprobs: An integer between 0 and 20 specifying the number of most
            likely tokens to return at each token position, each with an associated log
            probability. logprobs must be set to true if this parameter is used.

            max_completion_tokens: The maximum number of tokens that can be generated in
            the chat completion. The total length of input tokens and generated tokens
            is limited by the model's context length. Defaults to the provider's default
            max_completion_tokens when the value is None.

            n: How many chat completion choices to generate for each input message. Note
            that you will be charged based on the number of generated tokens across all
            of the choices. Keep n as 1 to minimize costs.

            presence_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on whether they appear in the text so far, increasing the
            model's likelihood to talk about new topics.

            response_format: An object specifying the format that the model must output.
            Setting to `{ "type": "json_schema", "json_schema": {...} }` enables
            Structured Outputs which ensures the model will match your supplied JSON
            schema. Learn more in the Structured Outputs guide. Setting to
            `{ "type": "json_object" }` enables JSON mode, which ensures the message the
            model generates is valid JSON.

            seed: If specified, a best effort attempt is made to sample
            deterministically, such that repeated requests with the same seed and
            parameters should return the same result. Determinism is not guaranteed, and
            you should refer to the system_fingerprint response parameter to monitor
            changes in the backend.

            stop: Up to 4 sequences where the API will stop generating further tokens.

            temperature:  What sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random,
            while lower values like 0.2 will make it more focused and deterministic.
            It is generally recommended to alter this or top_p, but not both.
            Defaults to the provider's default max_completion_tokens when the value is
            None.

            top_p: An alternative to sampling with temperature, called nucleus sampling,
            where the model considers the results of the tokens with top_p probability
            mass. So 0.1 means only the tokens comprising the top 10% probability mass
            are considered. Generally recommended to alter this or temperature, but not
            both.

            tools: A list of tools the model may call. Currently, only functions are
            supported as a tool. Use this to provide a list of functions the model may
            generate JSON inputs for. A max of 128 functions are supported.

            tool_choice: Controls which (if any) tool is called by the
            model. none means the model will not call any tool and instead generates a
            message. auto means the model can pick between generating a message or
            calling one or more tools. required means the model must call one or more
            tools. Specifying a particular tool via
            `{ "type": "function", "function": {"name": "my_function"} }`
            forces the model to call that tool.
            none is the default when no tools are present. auto is the default if tools
            are present.

            parallel_tool_calls: Whether to enable parallel function calling during tool
            use.

            use_custom_keys:  Whether to use custom API keys or our unified API keys
            with the backend provider.

            tags: Arbitrary number of tags to classify this API query as needed. Helpful
            for generally grouping queries across tasks and users, for logging purposes.

            drop_params: Whether or not to drop unsupported OpenAI params by the
            provider you’re using.

            region: A string used to represent the region where the endpoint is
            accessed. Only relevant for on-prem deployments with certain providers like
            `vertex-ai`, `aws-bedrock` and `azure-ml`, where the endpoint is being
            accessed through a specified region.

            log_query_body: Whether to log the contents of the query json body.

            log_response_body: Whether to log the contents of the response json body.

            stateful:  Whether the conversation history is preserved within the messages
            of this client. If True, then history is preserved. If False, then this acts
            as a stateless client, and message histories must be managed by the user.

            return_full_completion: If False, only return the message content
            chat_completion.choices[0].message.content.strip(" ") from the OpenAI
            return. Otherwise, the full response chat_completion is returned.
            Defaults to False.

            traced: Whether to trace the generate method.

            cache: If True, then the arguments will be stored in a local cache file, and
            any future calls with identical arguments will read from the cache instead
            of running the LLM query. If "write" then the cache will only be written
            to, if "read" then the cache will be read from if a cache is available but
            will not write, and if "read-only" then the argument must be present in the
            cache, else an exception will be raised. Finally, an appending "-closest"
            will read the closest match from the cache, and overwrite it if cache writing
            is enabled. This argument only has any effect when stream=False.

            extra_headers: Additional "passthrough" headers for the request which are
            provider-specific, and are not part of the OpenAI standard. They are handled
            by the provider-specific API.

            extra_query: Additional "passthrough" query parameters for the request which
            are provider-specific, and are not part of the OpenAI standard. They are
            handled by the provider-specific API.

            kwargs: Additional "passthrough" JSON properties for the body of the
            request, which are provider-specific, and are not part of the OpenAI
            standard. They will be handled by the provider-specific API.

        Raises:
            UnifyError: If the API key is missing.
        """
        self._base_constructor_args = dict(
            system_message=system_message,
            messages=messages,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            top_logprobs=top_logprobs,
            max_completion_tokens=max_completion_tokens,
            n=n,
            presence_penalty=presence_penalty,
            response_format=response_format,
            seed=seed,
            stop=stop,
            stream=False,
            stream_options=None,
            temperature=temperature,
            top_p=top_p,
            service_tier=service_tier,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            reasoning_effort=reasoning_effort,
            # platform arguments
            use_custom_keys=use_custom_keys,
            tags=tags,
            drop_params=drop_params,
            region=region,
            log_query_body=log_query_body,
            log_response_body=log_response_body,
            api_key=api_key,
            # python client arguments
            stateful=stateful,
            return_full_completion=return_full_completion,
            traced=traced,
            cache=cache,
            cache_backend=cache_backend,
            # passthrough arguments
            extra_headers=extra_headers,
            extra_query=extra_query,
            **kwargs,
        )
        super().__init__(**self._base_constructor_args)
        self._constructor_args = dict(
            endpoints=endpoints,
            **self._base_constructor_args,
        )
        if isinstance(endpoints, str):
            endpoints = [endpoints]
        else:
            endpoints = list(endpoints)
        self._api_key = _validate_api_key(api_key)
        self._endpoints = endpoints
        self._client_class = AsyncUnify
        self._clients = self._create_clients(endpoints)

    def _create_clients(self, endpoints: List[str]) -> Dict[str, AsyncUnify]:
        return {
            endpoint: self._client_class(
                endpoint,
                system_message=self.system_message,
                messages=self.messages,
                frequency_penalty=self.frequency_penalty,
                logit_bias=self.logit_bias,
                logprobs=self.logprobs,
                top_logprobs=self.top_logprobs,
                max_completion_tokens=self.max_completion_tokens,
                n=self.n,
                presence_penalty=self.presence_penalty,
                response_format=self.response_format,
                seed=self.seed,
                stop=self.stop,
                temperature=self.temperature,
                top_p=self.top_p,
                service_tier=self.service_tier,
                tools=self.tools,
                tool_choice=self.tool_choice,
                parallel_tool_calls=self.parallel_tool_calls,
                reasoning_effort=self.reasoning_effort,
                # platform arguments
                use_custom_keys=self.use_custom_keys,
                tags=self.tags,
                drop_params=self.drop_params,
                region=self.region,
                log_query_body=self.log_query_body,
                log_response_body=self.log_response_body,
                api_key=self._api_key,
                # python client arguments
                stateful=self.stateful,
                return_full_completion=self.return_full_completion,
                cache=self.cache,
                # passthrough arguments
                extra_headers=self.extra_headers,
                extra_query=self.extra_query,
                **self.extra_body,
            )
            for endpoint in endpoints
        }

    def add_endpoints(
        self,
        endpoints: Union[List[str], str],
        ignore_duplicates: bool = True,
    ) -> Self:
        """
        Add extra endpoints to be queried for each call to generate.

        Args:
            endpoints: The extra endpoints to add.

            ignore_duplicates: Whether or not to ignore duplicate endpoints passed.

        Returns:
            This client, useful for chaining inplace calls.
        """
        if isinstance(endpoints, str):
            endpoints = [endpoints]
        # remove duplicates
        if ignore_duplicates:
            endpoints = [
                endpoint for endpoint in endpoints if endpoint not in self._endpoints
            ]
        elif len(self._endpoints + endpoints) != len(set(self._endpoints + endpoints)):
            raise Exception(
                "at least one of the provided endpoints to add {}"
                "was already set present in the endpoints {}."
                "Set ignore_duplicates to True to ignore errors like this".format(
                    endpoints,
                    self._endpoints,
                ),
            )
        # update endpoints
        self._endpoints = self._endpoints + endpoints
        # create new clients
        self._clients.update(self._create_clients(endpoints))
        return self

    def remove_endpoints(
        self,
        endpoints: Union[List[str], str],
        ignore_missing: bool = True,
    ) -> Self:
        """
        Remove endpoints from the current list, which are queried for each call to
        generate.

        Args:
            endpoints: The extra endpoints to add.

            ignore_missing: Whether or not to ignore endpoints passed which are not
            currently present in the client endpoint list.

        Returns:
            This client, useful for chaining inplace calls.
        """
        if isinstance(endpoints, str):
            endpoints = [endpoints]
        # remove irrelevant
        if ignore_missing:
            endpoints = [
                endpoint for endpoint in endpoints if endpoint in self._endpoints
            ]
        elif len(self._endpoints) != len(set(self._endpoints + endpoints)):
            raise Exception(
                "at least one of the provided endpoints to remove {}"
                "was not present in the current endpoints {}."
                "Set ignore_missing to True to ignore errors like this".format(
                    endpoints,
                    self._endpoints,
                ),
            )
        # update endpoints and clients
        for endpoint in endpoints:
            self._endpoints.remove(endpoint)
            del self._clients[endpoint]
        return self

    def get_credit_balance(self) -> Union[float, None]:
        """
        Get the remaining credits left on your account.

        Returns:
            The remaining credits on the account if successful, otherwise None.
        Raises:
            BadRequestError: If there was an HTTP error.
            ValueError: If there was an error parsing the JSON response.
        """
        url = f"{BASE_URL}/credits"
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {self._api_key}",
        }
        try:
            response = _requests.get(url, headers=headers, timeout=10)
            if response.status_code != 200:
                raise Exception(response.json())
            return response.json()["credits"]
        except requests.RequestException as e:
            raise Exception("There was an error with the request.") from e
        except (KeyError, ValueError) as e:
            raise ValueError("Error parsing JSON response.") from e

    # Settable Properties #
    # --------------------#

    @property
    def endpoints(self) -> Tuple[str, ...]:
        """
        Get the current tuple of endpoints.

        Returns:
            The tuple of endpoints.
        """
        return tuple(self._endpoints)

    @property
    def clients(self) -> Dict[str, _UniClient]:
        """
        Get the current dictionary of clients, with endpoint names as keys and
        Unify or AsyncUnify instances as values.

        Returns:
            The dictionary of clients.
        """
        return self._clients

    # Representation #
    # ---------------#

    def __repr__(self):
        return "{}(endpoints={})".format(self.__class__.__name__, self._endpoints)

    def __str__(self):
        return "{}(endpoints={})".format(self.__class__.__name__, self._endpoints)

    # Generate #
    # ---------#

    def generate(
        self,
        arg0: Optional[Union[str, List[Union[str, Tuple[Any], Dict[str, Any]]]]] = None,
        /,
        system_message: Optional[str] = None,
        messages: Optional[
            Union[
                List[ChatCompletionMessageParam],
                Dict[str, List[ChatCompletionMessageParam]],
            ]
        ] = None,
        *,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, int]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        max_completion_tokens: Optional[int] = None,
        n: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        response_format: Optional[Union[Type[BaseModel], Dict[str, str]]] = None,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = None,
        stream: Optional[bool] = None,
        stream_options: Optional[ChatCompletionStreamOptionsParam] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        tools: Optional[Iterable[ChatCompletionToolParam]] = None,
        tool_choice: Optional[ChatCompletionToolChoiceOptionParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        reasoning_effort: Optional[str] = None,
        # platform arguments
        use_custom_keys: Optional[bool] = None,
        tags: Optional[List[str]] = None,
        drop_params: Optional[bool] = None,
        region: Optional[str] = None,
        log_query_body: Optional[bool] = None,
        log_response_body: Optional[bool] = None,
        # python client arguments
        stateful: Optional[bool] = None,
        return_full_completion: Optional[bool] = None,
        cache: Optional[Union[bool, str]] = None,
        # passthrough arguments
        extra_headers: Optional[Headers] = None,
        extra_query: Optional[Query] = None,
        **kwargs,
    ):
        """Generate a ChatCompletion response for the specified endpoint,
        from the provided query parameters.

        Args:
            arg0: A string containing the user message, or a list containing the inputs
            to send to each of the LLMs, in the format of str (user message), tuple
            (all-positional) or dict (all keyword).

            system_message: An optional string containing the system message. This
            always appears at the beginning of the list of messages.

            messages: A list of messages comprising the conversation so far, or
            optionally a dictionary of such messages, with clients as the keys in the
            case of multi-llm clients. This will be appended to the system_message if it
            is not None, and any user_message will be appended if it is not None.

            frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on their existing frequency in the text so far, decreasing the
            model's likelihood to repeat the same line verbatim.

            logit_bias: Modify the likelihood of specified tokens appearing in the
            completion. Accepts a JSON object that maps tokens (specified by their token
            ID in the tokenizer) to an associated bias value from -100 to 100.
            Mathematically, the bias is added to the logits generated by the model prior
            to sampling. The exact effect will vary per model, but values between -1 and
            1 should decrease or increase likelihood of selection; values like -100 or
            100 should result in a ban or exclusive selection of the relevant token.

            logprobs: Whether to return log probabilities of the output tokens or not.
            If true, returns the log probabilities of each output token returned in the
            content of message.

            top_logprobs: An integer between 0 and 20 specifying the number of most
            likely tokens to return at each token position, each with an associated log
            probability. logprobs must be set to true if this parameter is used.

            max_completion_tokens: The maximum number of tokens that can be generated in
            the chat completion. The total length of input tokens and generated tokens
            is limited by the model's context length. Defaults value is None. Uses the
            provider's default max_completion_tokens when None is explicitly passed.

            n: How many chat completion choices to generate for each input message. Note
            that you will be charged based on the number of generated tokens across all
            of the choices. Keep n as 1 to minimize costs.

            presence_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on whether they appear in the text so far, increasing the
            model's likelihood to talk about new topics.

            response_format: An object specifying the format that the model must output.
            Setting to `{ "type": "json_schema", "json_schema": {...} }` enables
            Structured Outputs which ensures the model will match your supplied JSON
            schema. Learn more in the Structured Outputs guide. Setting to
            `{ "type": "json_object" }` enables JSON mode, which ensures the message the
            model generates is valid JSON.

            seed: If specified, a best effort attempt is made to sample
            deterministically, such that repeated requests with the same seed and
            parameters should return the same result. Determinism is not guaranteed, and
            you should refer to the system_fingerprint response parameter to monitor
            changes in the backend.

            stop: Up to 4 sequences where the API will stop generating further tokens.

            stream: If True, generates content as a stream. If False, generates content
            as a single response. Defaults to False.

            stream_options: Options for streaming response. Only set this when you set
            stream: true.

            temperature:  What sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random,
            while lower values like 0.2 will make it more focused and deterministic.
            It is generally recommended to alter this or top_p, but not both.
            Default value is 1.0. Defaults to the provider's default temperature when
            None is explicitly passed.

            top_p: An alternative to sampling with temperature, called nucleus sampling,
            where the model considers the results of the tokens with top_p probability
            mass. So 0.1 means only the tokens comprising the top 10% probability mass
            are considered. Generally recommended to alter this or temperature, but not
            both.

            tools: A list of tools the model may call. Currently, only functions are
            supported as a tool. Use this to provide a list of functions the model may
            generate JSON inputs for. A max of 128 functions are supported.

            tool_choice: Controls which (if any) tool is called by the
            model. none means the model will not call any tool and instead generates a
            message. auto means the model can pick between generating a message or
            calling one or more tools. required means the model must call one or more
            tools. Specifying a particular tool via
            `{ "type": "function", "function": {"name": "my_function"} }`
            forces the model to call that tool.
            none is the default when no tools are present. auto is the default if tools
            are present.

            parallel_tool_calls: Whether to enable parallel function calling during tool
            use.

            stateful:  Whether the conversation history is preserved within the messages
            of this client. If True, then history is preserved. If False, then this acts
            as a stateless client, and message histories must be managed by the user.

            use_custom_keys:  Whether to use custom API keys or our unified API keys
            with the backend provider. Defaults to False.

            tags: Arbitrary number of tags to classify this API query as needed. Helpful
            for generally grouping queries across tasks and users, for logging purposes.

            drop_params: Whether or not to drop unsupported OpenAI params by the
            provider you’re using.

            region: A string used to represent the region where the endpoint is
            accessed. Only relevant for on-prem deployments with certain providers like
            `vertex-ai`, `aws-bedrock` and `azure-ml`, where the endpoint is being
            accessed through a specified region.

            log_query_body: Whether to log the contents of the query json body.

            log_response_body: Whether to log the contents of the response json body.

            stateful:  Whether the conversation history is preserved within the messages
            of this client. If True, then history is preserved. If False, then this acts
            as a stateless client, and message histories must be managed by the user.

            return_full_completion: If False, only return the message content
            chat_completion.choices[0].message.content.strip(" ") from the OpenAI
            return. Otherwise, the full response chat_completion is returned.
            Defaults to False.

            cache: If True, then the arguments will be stored in a local cache file, and
            any future calls with identical arguments will read from the cache instead
            of running the LLM query. If "write" then the cache will only be written
            to, if "read" then the cache will be read from if a cache is available but
            will not write, and if "read-only" then the argument must be present in the
            cache, else an exception will be raised. Finally, an appending "-closest"
            will read the closest match from the cache, and overwrite it if cache writing
            is enabled. This argument only has any effect when stream=False.

            extra_headers: Additional "passthrough" headers for the request which are
            provider-specific, and are not part of the OpenAI standard. They are handled
            by the provider-specific API.

            extra_query: Additional "passthrough" query parameters for the request which
            are provider-specific, and are not part of the OpenAI standard. They are
            handled by the provider-specific API.

            kwargs: Additional "passthrough" JSON properties for the body of the
            request, which are provider-specific, and are not part of the OpenAI
            standard. They will be handled by the provider-specific API.

        Returns:
            If stream is True, returns a generator yielding chunks of content.
            If stream is False, returns a single string response.

        Raises:
            UnifyError: If an error occurs during content generation.
        """
        system_message = _default(system_message, self._system_message)
        messages = _default(messages, self._messages)
        stateful = _default(stateful, self._stateful)
        if messages:
            # system message only added once at the beginning
            if isinstance(arg0, str):
                if isinstance(messages, dict):
                    messages = {
                        k: v + [{"role": "user", "content": arg0}]
                        for k, v in messages.items()
                    }
                else:
                    messages += [{"role": "user", "content": arg0}]
        else:
            messages = list()
            if system_message is not None:
                messages += [{"role": "system", "content": system_message}]
            if isinstance(arg0, str):
                messages += [{"role": "user", "content": arg0}]
            self._messages = messages
        return_full_completion = (
            True
            if _default(tools, self._tools)
            else _default(return_full_completion, self._return_full_completion)
        )
        ret = self._generate(
            messages=messages,
            frequency_penalty=_default(frequency_penalty, self._frequency_penalty),
            logit_bias=_default(logit_bias, self._logit_bias),
            logprobs=_default(logprobs, self._logprobs),
            top_logprobs=_default(top_logprobs, self._top_logprobs),
            max_completion_tokens=_default(
                max_completion_tokens,
                self._max_completion_tokens,
            ),
            n=_default(n, self._n),
            presence_penalty=_default(presence_penalty, self._presence_penalty),
            response_format=_default(response_format, self._response_format),
            seed=_default(_default(seed, self._seed), unify.get_seed()),
            stop=_default(stop, self._stop),
            stream=_default(stream, self._stream),
            stream_options=_default(stream_options, self._stream_options),
            temperature=_default(temperature, self._temperature),
            top_p=_default(top_p, self._top_p),
            tools=_default(tools, self._tools),
            tool_choice=_default(tool_choice, self._tool_choice),
            parallel_tool_calls=_default(
                parallel_tool_calls,
                self._parallel_tool_calls,
            ),
            reasoning_effort=_default(reasoning_effort, self._reasoning_effort),
            # platform arguments
            use_custom_keys=_default(use_custom_keys, self._use_custom_keys),
            tags=_default(tags, self._tags),
            drop_params=_default(drop_params, self._drop_params),
            region=_default(region, self._region),
            log_query_body=_default(log_query_body, self._log_query_body),
            log_response_body=_default(log_response_body, self._log_response_body),
            # python client arguments
            return_full_completion=return_full_completion,
            cache=_default(cache, self._cache),
            # passthrough arguments
            extra_headers=_default(extra_headers, self._extra_headers),
            extra_query=_default(extra_query, self._extra_query),
            **{**self._extra_body, **kwargs},
        )
        if stateful:
            if return_full_completion:
                msg = [ret.choices[0].message.model_dump()]
            else:
                msg = [{"role": "assistant", "content": ret}]
            if self._messages is None:
                self._messages = []
            self._messages += msg
        return ret


class MultiUnify(_MultiClient):
    async def _async_gen(
        self,
        messages: Optional[
            Union[
                List[ChatCompletionMessageParam],
                Dict[str, List[ChatCompletionMessageParam]],
            ]
        ] = None,
        *,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, int]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        max_completion_tokens: Optional[int] = None,
        n: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        response_format: Optional[Union[Type[BaseModel], Dict[str, str]]] = None,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = None,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = None,
        tools: Optional[Iterable[ChatCompletionToolParam]] = None,
        tool_choice: Optional[ChatCompletionToolChoiceOptionParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        reasoning_effort: Optional[str] = None,
        # platform arguments
        use_custom_keys: bool = False,
        tags: Optional[List[str]] = None,
        drop_params: Optional[bool] = True,
        region: Optional[str] = None,
        log_query_body: Optional[bool] = True,
        log_response_body: Optional[bool] = True,
        # python client arguments
        return_full_completion: bool = False,
        cache_backend: Optional[str] = None,
        # passthrough arguments
        extra_headers: Optional[Headers] = None,
        extra_query: Optional[Query] = None,
        service_tier: Optional[str] = None,
        **kwargs,
    ) -> Union[Union[str, ChatCompletion], Dict[str, Union[str, ChatCompletion]]]:
        kw = dict(
            messages=messages,
            max_completion_tokens=max_completion_tokens,
            stop=stop,
            temperature=temperature,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            top_logprobs=top_logprobs,
            n=n,
            presence_penalty=presence_penalty,
            response_format=response_format,
            seed=seed,
            top_p=top_p,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            reasoning_effort=reasoning_effort,
            use_custom_keys=use_custom_keys,
            tags=tags,
            drop_params=drop_params,
            region=region,
            log_query_body=log_query_body,
            log_response_body=log_response_body,
            return_full_completion=return_full_completion,
            extra_headers=extra_headers,
            extra_query=extra_query,
            service_tier=service_tier,
            **kwargs,
        )
        multi_message = isinstance(kw["messages"], dict)
        kw_ = {k: v for k, v in kw.items() if v is not None}
        responses = dict()
        for endpoint, client in self._clients.items():
            these_kw = kw_.copy()
            if multi_message:
                these_kw["messages"] = these_kw["messages"][endpoint]
            responses[endpoint] = await client.generate(**these_kw)
        return responses[self._endpoints[0]] if len(self._endpoints) == 1 else responses

    async def _multi_inp_gen(
        self,
        multi_input: List[Union[str, Tuple[Any], Dict[str, Any]]],
        **kwargs,
    ) -> List[Union[str, ChatCompletion]]:
        assert isinstance(multi_input, list), (
            f"Expected multi_kwargs to be a list, "
            f"but found {multi_input} of type {type(multi_input)}."
        )
        assert all(
            type(multi_input[0]) is type(i) for i in multi_input
        ), "all entries in the list of inputs must be of the same type."
        if isinstance(multi_input[0], str):
            coroutines = [self._async_gen(s, **kwargs) for s in multi_input]
        elif isinstance(multi_input[0], tuple):
            coroutines = [self._async_gen(*a, **kwargs) for a in multi_input]
        elif isinstance(multi_input[0], dict):
            coroutines = [self._async_gen(**{**kwargs, **kw}) for kw in multi_input]
        else:
            raise Exception(
                f"Invalid format for first argument in list, expected either str, "
                f"tuple or dict but found "
                f"{multi_input[0]} of type {type(multi_input[0])}.",
            )
        return await asyncio.gather(*coroutines)

    def _multi_inp_generate(
        self,
        *args,
        **kwargs,
    ) -> List[Union[str, ChatCompletion]]:
        """
        Perform multiple generations to multiple inputs asynchronously, based on the
        list keywords arguments passed in.
        """
        return asyncio.run(self._multi_inp_gen(*args, **kwargs))

    def _generate(  # noqa: WPS234, WPS211
        self,
        *args,
        **kwargs,
    ) -> Union[
        Union[
            Union[str, ChatCompletion],
            List[Union[str, ChatCompletion]],
        ],
        Dict[
            str,
            Union[
                Union[str, ChatCompletion],
                List[Union[str, ChatCompletion]],
            ],
        ],
    ]:
        # refresh the openai client before doing a new event loop
        for key, client in self._clients.items():
            client._client = client._get_client()
        if args and isinstance(args[0], list):
            return self._multi_inp_generate(*args, **kwargs)
        return asyncio.run(
            self._async_gen(
                *args,
                **kwargs,
            ),
        )

    def to_async_client(self):
        """
        Return an asynchronous version of the client (`AsyncMultiUnify` instance), with
        the exact same configuration as this synchronous (`MultiUnify`) client.

        Returns:
            An `AsyncMultiUnify` instance with the same configuration as this `MultiUnify`
            instance.
        """
        return AsyncMultiUnify(**self._constructor_args)


class AsyncMultiUnify(_MultiClient):
    async def _generate(  # noqa: WPS234, WPS211
        self,
        messages: Optional[
            Union[
                List[ChatCompletionMessageParam],
                Dict[str, List[ChatCompletionMessageParam]],
            ]
        ] = None,
        *,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, int]] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        max_completion_tokens: Optional[int] = None,
        n: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        response_format: Optional[Union[Type[BaseModel], Dict[str, str]]] = None,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = None,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = None,
        tools: Optional[Iterable[ChatCompletionToolParam]] = None,
        tool_choice: Optional[ChatCompletionToolChoiceOptionParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        reasoning_effort: Optional[str] = None,
        # platform arguments
        use_custom_keys: bool = False,
        tags: Optional[List[str]] = None,
        drop_params: Optional[bool] = True,
        region: Optional[str] = None,
        log_query_body: Optional[bool] = True,
        log_response_body: Optional[bool] = True,
        # python client arguments
        return_full_completion: bool = False,
        cache_backend: Optional[str] = None,
        # passthrough arguments
        extra_headers: Optional[Headers] = None,
        extra_query: Optional[Query] = None,
        service_tier: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, str]:
        kw = dict(
            messages=messages,
            max_completion_tokens=max_completion_tokens,
            stop=stop,
            temperature=temperature,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            top_logprobs=top_logprobs,
            n=n,
            presence_penalty=presence_penalty,
            response_format=response_format,
            seed=seed,
            top_p=top_p,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            reasoning_effort=reasoning_effort,
            use_custom_keys=use_custom_keys,
            tags=tags,
            drop_params=drop_params,
            region=region,
            log_query_body=log_query_body,
            log_response_body=log_response_body,
            return_full_completion=return_full_completion,
            cache_backend=cache_backend,
            extra_headers=extra_headers,
            extra_query=extra_query,
            service_tier=service_tier,
            **kwargs,
        )
        multi_message = isinstance(messages, dict)
        kw = {k: v for k, v in kw.items() if v is not None}
        responses = dict()
        for endpoint, client in self._clients.items():
            these_kw = kw.copy()
            if multi_message:
                these_kw["messages"] = these_kw["messages"][endpoint]
            responses[endpoint] = await client.generate(**these_kw)
        return responses

    def to_sync_client(self):
        """
        Return a synchronous version of the client (`MultiUnify` instance), with the
        exact same configuration as this asynchronous (`AsyncMultiUnify`) client.

        Returns:
            A `MultiUnify` instance with the same configuration as this `AsyncMultiUnify`
            instance.
        """
        return MultiUnify(**self._constructor_args)



import functools
import os
import warnings
from dataclasses import dataclass


@dataclass
class CacheStats:
    hits: int = 0
    misses: int = 0
    reads: int = 0
    writes: int = 0

    def get_percentage_of_cache_hits(self) -> float:
        if self.reads == 0:
            return 0.0
        return self.hits / self.reads * 100

    def get_percentage_of_cache_misses(self) -> float:
        if self.reads == 0:
            return 0.0
        return self.misses / self.reads * 100

    def __repr__(self) -> str:
        return f"CacheStats(hits={self.hits} ({self.get_percentage_of_cache_hits():.1f}%), misses={self.misses} ({self.get_percentage_of_cache_misses():.1f}%), reads={self.reads}, writes={self.writes})"


CURRENT_CACHE_STATS = CacheStats()


def _is_cache_benchmark_enabled() -> bool:
    return os.environ.get("UNIFY_CACHE_BENCHMARK", "false") == "true"


def get_cache_stats() -> CacheStats:
    if not _is_cache_benchmark_enabled():
        warnings.warn(
            "Cache benchmark is not enabled, set UNIFY_CACHE_BENCHMARK=true to enable it, must be set before importing unify.",
        )
    return CURRENT_CACHE_STATS


def reset_cache_stats() -> None:
    global CURRENT_CACHE_STATS
    CURRENT_CACHE_STATS = CacheStats()


def record_get_cache(fn):
    if not _is_cache_benchmark_enabled():
        return fn
    else:

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            benchmark = get_cache_stats()
            benchmark.reads += 1
            ret = fn(*args, **kwargs)
            if ret is None:
                benchmark.misses += 1
            else:
                benchmark.hits += 1
            return ret

        return wrapper


def record_write_to_cache(fn):
    if not _is_cache_benchmark_enabled():
        return fn
    else:

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            benchmark = get_cache_stats()
            benchmark.writes += 1
            return fn(*args, **kwargs)

        return wrapper



"""
Local file-based cache implementation.

This cache stores data in a local JSON file and provides fast in-memory access.
"""

import json
import os
import threading
import warnings
from typing import Any, Dict, List, Optional

from unify.utils.caching.base_cache import BaseCache


class LocalCache(BaseCache):
    """Local file-based cache implementation."""

    _cache: Optional[Dict[str, Any]] = None
    _cache_dir: str = os.environ.get("UNIFY_CACHE_DIR", os.getcwd())
    _cache_lock: threading.Lock = threading.Lock()
    _cache_filename: str = ".cache.json"
    _enabled: bool = False

    @classmethod
    def set_cache_name(cls, name: str) -> None:
        """Set the cache filename and reset the in-memory cache."""
        cls._cache_filename = name
        cls._cache = None  # Force reload on next access

    @classmethod
    def get_cache_name(cls) -> str:
        """Get the current cache filename."""
        return cls._cache_filename

    @classmethod
    def get_cache_filepath(cls, name: str = None) -> str:
        """Get the full filepath for the cache file."""
        if name is None:
            name = cls.get_cache_name()
        return os.path.join(cls._cache_dir, name)

    @classmethod
    def is_enabled(cls) -> bool:
        """Check if the cache is enabled."""
        return cls._enabled

    @classmethod
    def store_entry(
        cls,
        *,
        key: str,
        value: Any,
        res_types: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Store a key-value pair in the cache."""
        if res_types:
            cls._cache[f"{key}_res_types"] = res_types
        cls._cache[key] = value
        cls.write()

    @classmethod
    def write(cls, filename: str = None) -> None:
        """Write the cache to disk."""
        cache_filepath = cls.get_cache_filepath(filename)
        with open(cache_filepath, "w") as outfile:
            json.dump(cls._cache, outfile, indent=2)

    @classmethod
    def initialize_cache(cls, name: str = None) -> None:
        """Initialize or load the cache from disk."""
        cache_filepath = cls.get_cache_filepath(name)

        if cls._cache is None:
            # Create cache file if it doesn't exist
            if not os.path.exists(cache_filepath):
                with open(cache_filepath, "w") as outfile:
                    json.dump({}, outfile)
                cls._cache = {}
                return

            # Load existing cache
            try:
                with open(cache_filepath, "r") as infile:
                    content = infile.read().strip()
                    if not content:
                        # File is empty, initialize with empty dict
                        cls._cache = {}
                    else:
                        # Parse the JSON content
                        cls._cache = json.loads(content)
            except (json.JSONDecodeError, IOError):
                # File contains invalid JSON or can't be read, reinitialize
                warnings.warn(
                    f"Cache file {cache_filepath} contains invalid JSON or can't be read, reinitializing.",
                )
                cls._cache = {}

    @classmethod
    def list_keys(cls) -> List[str]:
        return list(cls._cache.keys())

    @classmethod
    def retrieve_entry(cls, key: str) -> tuple[Optional[Any], Optional[Dict[str, Any]]]:
        """
        Retrieve a value from the cache.

        Returns:
            Tuple of (value, type_registry) or (None, None) if not found
        """
        if cls._cache is None:
            return None, None

        value = cls._cache.get(key)
        if value is None:
            return None, None

        deserialized_value = json.loads(value)
        res_types = cls._cache.get(f"{key}_res_types")
        return deserialized_value, res_types

    @classmethod
    def has_key(cls, key: str) -> bool:
        """Check if a key exists in the cache."""
        return cls._cache is not None and key in cls._cache

    @classmethod
    def remove_entry(cls, key: str) -> None:
        """Remove an entry and its res_types from the cache."""
        if cls._cache is not None:
            del cls._cache[key]
            del cls._cache[f"{key}_res_types"]
            cls.write()



from __future__ import annotations

import json
from collections.abc import Sequence
from typing import Any, Dict, List, Optional, Union

import unify
from typing_extensions import Self

from ..universal_api.types import Prompt

# noinspection PyProtectedMember
from ..utils.helpers import _validate_api_key


def _to_raw_data(x: Dict[str, Any]):
    return x["data"] if "data" in x and len(x) == 1 else x


class Dataset(Sequence):
    def __init__(
        self,
        data: Optional[Any] = None,
        *,
        name: str = None,
        allow_duplicates: bool = False,
        api_key: Optional[str] = None,
    ) -> None:
        """
        Initialize a local dataset.

        Args:
            data: The data for populating the dataset.
            This needs to be a list of JSON serializable objects.

            name: The name of the dataset. To create a dataset for a specific project
            with name {project_name}, then prefix the name with {project_name}/{name}.

            allow_duplicates: Whether to allow duplicates in the dataset.

            api_key: API key for accessing the Unify API. If None, it attempts to
            retrieve the API key from the environment variable UNIFY_KEY. Defaults to
            None.

        Raises:
            UnifyError: If the API key is missing.
        """
        self._name = name
        if isinstance(data, tuple):
            data = list(data)
        elif not isinstance(data, list):
            data = [data]
        self._allow_duplicates = allow_duplicates
        self._api_key = _validate_api_key(api_key)
        self._logs = [
            (
                entry
                if isinstance(entry, unify.Log)
                else unify.Log(
                    **(entry if isinstance(entry, dict) else {"data": entry}),
                )
            )
            for entry in data
        ]
        super().__init__()

    @property
    def name(self) -> str:
        """
        Name of the dataset.
        """
        return self._name

    @property
    def allow_duplicates(self) -> bool:
        """
        Whether to allow duplicates in the dataset.
        """
        return self._allow_duplicates

    @property
    def data(self):
        """
        Dataset entries.
        """
        return [_to_raw_data(l.entries) for l in self._logs]

    def _set_data(self, data):
        self._logs = [
            unify.Log(**(entry if isinstance(entry, dict) else {"data": entry}))
            for entry in data
        ]

    def set_name(self, name: str) -> Self:
        """
        Set the name of the dataset.

        Args:
            name: The name to set the dataset to.

        Returns:
            This dataset, useful for chaining methods.
        """
        self._name = name
        return self

    def set_allow_duplicates(self, allow_duplicates: bool) -> Self:
        """
        Set whether to allow duplicates in the dataset.

        Args:
            allow_duplicates: Whether to allow duplicates in the dataset.

        Returns:
            This dataset, useful for chaining methods.
        """
        self._allow_duplicates = allow_duplicates
        return self

    @staticmethod
    def from_upstream(
        name: str,
        api_key: Optional[str] = None,
    ) -> Dataset:
        """
        Initialize a local dataset from the upstream dataset.

        Args:
            name: The name of the dataset.

            api_key: API key for accessing the Unify API. If None, it attempts to
            retrieve the API key from the environment variable UNIFY_KEY. Defaults to
            None.

        Returns:
            The dataset, with contents downloaded from upstream.

        Raises:
            UnifyError: If the API key is missing.
        """
        data = unify.download_dataset(name=name, api_key=api_key)
        return Dataset(
            data,
            name=name,
            api_key=api_key,
        )

    def _assert_name_exists(self) -> None:
        assert self._name is not None, (
            "Dataset name must be specified in order to upload, download, sync or "
            "compare to a corresponding dataset in your upstream account. "
            "You can simply use .set_name() and set it to the same name as your "
            "upstream dataset, or create a new name if it doesn't yet exist upstream."
        )

    def upload(self, overwrite: bool = False) -> Self:
        """
        Uploads all unique local data in the dataset to the user account upstream.
        This function will not download any uniques from upstream.
        Use `sync` to synchronize and superset the datasets in both directions.
        Set `overwrite=True` to disregard any pre-existing upstream data.

        Args:
            overwrite: Whether to overwrite the upstream dataset if it already exists.

        Returns:
            This dataset, useful for chaining methods.
        """
        self._assert_name_exists()
        dataset_ids = unify.upload_dataset(
            name=self._name,
            data=self._logs,
            overwrite=overwrite,
            allow_duplicates=self._allow_duplicates,
        )
        assert len(dataset_ids) >= len(
            self._logs,
        ), "Number of upstream items must be greater than or equal to items"
        return self

    def download(self, overwrite: bool = False) -> Self:
        """
        Downloads all unique upstream data from the user account to the local dataset.
        This function will not upload any unique values stored locally.
        Use `sync` to synchronize and superset the datasets in both directions.
        Set `overwrite=True` to disregard any pre-existing data stored in this class.

        Args:
            overwrite: Whether to overwrite the local data, if any already exists

        Returns:
            This dataset after the in-place download, useful for chaining methods.
        """
        self._assert_name_exists()

        if f"Datasets/{self._name}" not in unify.get_contexts():
            upstream_dataset = list()
        else:
            upstream_dataset = unify.download_dataset(
                name=self._name,
                api_key=self._api_key,
            )
        if overwrite:
            self._logs = upstream_dataset
            return self
        if self._allow_duplicates:
            local_ids = set([l.id for l in self._logs if l.id is not None])
            new_data = [l for l in upstream_dataset if l.id not in local_ids]
        else:
            local_vals_to_logs = {json.dumps(l.entries): l for l in self._logs}
            local_values = set([json.dumps(l.entries) for l in self._logs])
            upstream_values = set()
            new_data = list()
            for l in upstream_dataset:
                uid = l.id
                value = json.dumps(l.entries)
                if value not in local_values.union(upstream_values):
                    new_data.append(l)
                    upstream_values.add(value)
                elif value in local_vals_to_logs:
                    local_log = local_vals_to_logs[value]
                    if local_log.id != uid:
                        local_log.set_id(uid)

        self._logs += new_data
        return self

    def sync(self) -> Self:
        """
        Synchronize the dataset in both directions, downloading any values missing
        locally, and uploading any values missing from upstream in the account.

        Returns:
            This dataset after the in-place sync, useful for chaining methods.
        """
        self.upload()
        self.download(overwrite=True)
        return self

    def upstream_diff(self) -> Self:
        """
        Prints the difference between the local dataset and the upstream dataset.

        Returns:
            This dataset after printing the diff, useful for chaining methods.
        """
        self._assert_name_exists()
        upstream_dataset = unify.download_dataset(
            name=self._name,
            api_key=self._api_key,
        )
        unique_upstream = [
            item["entry"] for item in upstream_dataset if item["entry"] not in self.data
        ]
        print(
            "The following {} entries are stored upstream but not locally\n: "
            "{}".format(len(unique_upstream), unique_upstream),
        )
        unique_local = [item for item in self.data if item not in upstream_dataset]
        print(
            "The following {} entries are stored upstream but not locally\n: "
            "{}".format(len(unique_local), unique_local),
        )
        return self

    def add(
        self,
        other: Union[
            Dataset,
            str,
            Dict,
            Prompt,
            int,
            List[Union[str, Dict, Prompt]],
        ],
    ) -> Self:
        """
        Adds another dataset to this one, return a new Dataset instance, with this
        new dataset receiving all unique queries from the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the addition.
        """
        if other == 0:
            return self
        other = other if isinstance(other, Dataset) else Dataset(other)
        data = self.data + [d for d in other.data if d not in self.data]
        return Dataset(data=data, api_key=self._api_key)

    def sub(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Subtracts another dataset from this one, return a new Dataset instance, with
        this new dataset losing all queries from the other subtracted dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the subtraction.
        """
        other = other if isinstance(other, Dataset) else Dataset(other)
        assert other in self, (
            "cannot subtract dataset B from dataset A unless all queries of dataset "
            "B are also present in dataset A"
        )
        data = [item for item in self.data if item not in other]
        return Dataset(data=data, api_key=self._api_key)

    def inplace_add(
        self,
        other: Union[
            Dataset,
            str,
            Dict,
            Prompt,
            int,
            List[Union[str, Dict, Prompt]],
        ],
    ) -> Self:
        """
        Adds another dataset to this one, with this dataset receiving all unique queries
        from the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place addition.
        """
        if other == 0:
            return self
        other = other if isinstance(other, Dataset) else Dataset(other)
        self._logs = self._logs + [d for d in other._logs if d not in self._logs]
        return self

    def inplace_sub(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Subtracts another dataset from this one, with this dataset losing all queries
        from the other subtracted dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place subtraction.
        """
        other = other if isinstance(other, Dataset) else Dataset(other)
        assert other in self, (
            "cannot subtract dataset B from dataset A unless all queries of dataset "
            "B are also present in dataset A"
        )
        self._logs = [item for item in self._logs if item not in other]
        return self

    def __add__(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Adds another dataset to this one via the + operator, return a new Dataset
        instance, with this new dataset receiving all unique queries from the other
        added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the addition.
        """
        return self.add(other)

    def __radd__(
        self,
        other: Union[
            Dataset,
            str,
            Dict,
            Prompt,
            int,
            List[Union[str, Dict, Prompt]],
        ],
    ) -> Self:
        """
        Adds another dataset to this one via the + operator, this is used if the
        other item does not have a valid __add__ method for these two types. Return a
        new Dataset instance, with this new dataset receiving all unique queries from
        the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            The new dataset following the addition.
        """
        if other == 0:
            return self
        return Dataset(other).add(self)

    def __iadd__(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Adds another dataset to this one, with this dataset receiving all unique queries
        from the other added dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place addition.
        """
        return self.inplace_add(other)

    def __sub__(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Subtracts another dataset from this one via the - operator, return a new Dataset
        instance, with this new dataset losing all queries from the other subtracted
        dataset.

        Args:
            other: The other dataset being subtracted from this one.

        Returns:
            The new dataset following the subtraction.
        """
        return self.sub(other)

    def __rsub__(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Subtracts another dataset from this one via the - operator, this is used if the
        other item does not have a valid __sub__ method for these two types. Return a
        new Dataset instance, with this new dataset losing all queries from the other
        subtracted dataset.

        Args:
            other: The other dataset being subtracted from this one.

        Returns:
            The new dataset following the subtraction.
        """
        return Dataset(other).sub(self)

    def __isub__(
        self,
        other: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> Self:
        """
        Subtracts another dataset from this one, with this dataset losing all queries
        from the other subtracted dataset.

        Args:
            other: The other dataset being added to this one.

        Returns:
            This dataset following the in-place subtraction.
        """
        return self.inplace_sub(other)

    def __iter__(self) -> Any:
        """
        Iterates through the dataset, return one instance at a time.

        Returns:
            The next instance in the dataset.
        """
        for l in self._logs:
            yield l

    def __contains__(
        self,
        item: Union[Dataset, str, Dict, Prompt, List[Union[str, Dict, Prompt]]],
    ) -> bool:
        """
        Determine whether the item is contained within the dataset. The item is cast to
        a Dataset instance, and can therefore take on many different types. Only returns
        True if *all* entries in the passed dataset are contained within this dataset.

        Args:
            item: The item to cast to a Dataset before checking if it's a subset of this
            one.

        Returns:
            Boolean, whether the passed Dataset is a subset of this one.
        """
        item = item if isinstance(item, Dataset) else Dataset(item)
        this_serialized = [
            json.dumps(
                {
                    k: v
                    for k, v in l.to_json().items()
                    if k not in ("id", "ts") and v not in ({}, None)
                },
            )
            for l in self._logs
        ]
        item_serialized = [
            json.dumps(
                {
                    k: v
                    for k, v in l.to_json().items()
                    if k not in ("id", "ts") and v not in ({}, None)
                },
            )
            for l in item._logs
        ]
        this_set = set(this_serialized)
        combined_set = set(this_serialized + item_serialized)
        return len(this_set) == len(combined_set)

    def __len__(self) -> int:
        """
        Returns the number of entries contained within the dataset.

        Returns:
            The number of entries in the dataset.
        """
        return len(self._logs)

    def __getitem__(self, item: Union[int, slice]) -> Union[Any, Dataset]:
        """
        Gets an item from the dataset, either via an int or slice. In the case of an
        int, then a data instance is returned, and for a slice a Dataset instance is
        returned.

        Args:
            item: integer or slice for extraction.

        Returns:
            An individual item or Dataset slice, for int and slice queries respectively.
        """
        if isinstance(item, int):
            return self._logs[item]
        elif isinstance(item, slice):
            return Dataset(self._logs[item.start : item.stop : item.step])
        raise TypeError(
            "expected item to be of type int or slice,"
            "but found {} of type {}".format(item, type(item)),
        )

    def __setitem__(self, item: Union[int, slice], value: Union[Any, Dataset]):
        if isinstance(item, int):
            if isinstance(value, unify.Log):
                self._logs[item] = value
            else:
                self._logs[item] = unify.Log(
                    **(value if isinstance(value, dict) else {"data": value}),
                )
        elif isinstance(item, slice):
            self._logs[item.start : item.stop : item.step] = [
                (
                    unify.Log(**(v if isinstance(v, dict) else {"data": v}))
                    if not isinstance(v, unify.Log)
                    else v
                )
                for v in value
            ]
        else:
            raise TypeError(
                "expected item to be of type int or slice,",
            )

    def __repr__(self):
        return f"unify.Dataset({self.data}, name='{self._name}')"
